import os
import json


def load_json(path):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    file_path = os.path.join(dir_path, path)
    with open(file_path, "r") as fp:
        return json.load(fp)


def load_content(path):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    file_path = os.path.join(dir_path, path)
    with open(file_path, "rb") as fp:
        return fp.read()



import re
import pandas as pd
from uuid import uuid4

from datetime import date
from dateutil.relativedelta import relativedelta

from myotest.wrapper import (
    WrapperObject, StringField, DateTimeField, ListField,
    IDField, FloatField, DurationField, DataframeField, LinkField,
    DictField, DateField, IntegerField, BooleanField
)

from myotest.exception import ResourceNotFoundException


class AvroSchemaField(WrapperObject):
    fields = ListField(DictField())
    name = StringField()
    type = StringField()


class Dataset(WrapperObject):
    name = StringField()
    dataframe = DataframeField()
    workout = LinkField()
    describe = DictField()
    avro_schema = AvroSchemaField()

    def resolve_dataframe(self, json, field):
        reader = self.client.fetch_avro(json["cloud_url"])
        return pd.DataFrame(list(reader)).sort_values("time")

    def dataframe_for_slot(self, slot):
        start_time = slot.start_time.total_seconds()
        end_time = slot.end_time.total_seconds()
        df = self.dataframe
        return df[
            (df["time"] < end_time) &
            (df["time"] >= start_time)
        ]

    @property
    def type(self):
        return self.name.split("-")[0]


class SlotValue(WrapperObject):
    type = StringField()
    value = FloatField()


class ResultData(WrapperObject):
    min = FloatField()
    max = FloatField()
    count = IntegerField()
    mean = FloatField()
    median = FloatField()
    std = FloatField()


class SlotResult(WrapperObject):
    power = ResultData()
    speed = ResultData()
    cadence = ResultData()
    undulation = ResultData()
    stiffness = ResultData()
    stride_length = ResultData()
    effective_flight_time = ResultData()
    effective_contact_time = ResultData()
    distance = FloatField()
    regularity_90 = FloatField()
    step_count_ratio = FloatField()
    regularity_median = FloatField()
    gps_valid = BooleanField()
    cmj_note = FloatField()


class Slot(WrapperObject):
    id = IDField()
    tags = ListField(StringField())
    type = StringField()
    value = SlotValue()
    text = StringField()
    result = SlotResult()
    end_time = DurationField()
    start_time = DurationField()
    power_type = StringField()
    analysis = ListField(DictField())
    workout = LinkField()

    def resolve_dataframe(self, json, field):
        reader = self.client.fetch_avro(json["cloud_url"])
        return pd.DataFrame(list(reader)).sort_values("time")

    def get_dataframe(self, dataset_name):
        return self.workout.get_dataset(dataset_name).dataframe_for_slot(self)

    @property
    def duration(self):
        return self.end_time - self.start_time


class Workout(WrapperObject):
    id = IDField()
    title = StringField()
    start = DateTimeField()
    end = DateTimeField()
    type = StringField()
    target_duration = DurationField()
    effective_duration = DurationField()
    tags = ListField(StringField())

    datasets = ListField(Dataset(), source="data")
    slots = ListField(Slot())

    def _get_datasets(self, base_name):
        if "-" in base_name:
            regexp = re.compile("^{}$".format(base_name))
        else:
            regexp = re.compile("^{}-[0-9]+$".format(base_name))
        return [x for x in self.datasets if re.match(regexp, x.name)]

    def post_resolve_datasets(self, datasets):
        for ds in datasets:
            ds.workout = self

    def get_datasets(self, name):
        return list(self._get_datasets(name))

    def get_dataset(self, name):
        datasets = self._get_datasets(name)
        if len(datasets) > 0:
            return datasets[0]
        else:
            return None

    def resolve_slots(self, json, field):
        try:
            slots = self.client.get_slots(json["id"])
        except ResourceNotFoundException:
            slots = None
        if slots is None:
            slots = self.create_fake_slots()
        return slots

    def create_fake_slots(self):
        """
        Create a single slot matching the full workout
        This is for legacy workout without validation and slots
        :return: Slot
        """
        single_slot = {
            "id": str(uuid4()),
            "text": self.title,
            "tags": self.tags,
            "type": "unknown",
            "value": {
                "type": "duration",
                "value": self.target_duration.total_seconds(),
            },
            "end_time": self.effective_duration.total_seconds(),
            "power_type": "",
            "start_time": 0,
            "result": {
                "regularity_90": 1.0,
                "regularity_median": 1.0,
                "step_count_ratio": 0.0,
                "speed": None,
                "distance": 0,
                "power": None,
                "cadence": None,
                "undulation": None,
                "stiffness": None,
                "stride_length": None,
                "effective_flight_time": None,
                "effective_contact_time": None,
                "gps_valid": None,
                "cmj_note": 0
            }
        }

        mil_ds = self.get_dataset("mil")
        if mil_ds:
            rounding = 3

            def extract_result(key):
                if key not in mil_ds.describe:
                    return {
                        "max": 0,
                        "min": 0,
                        "std": 0,
                        "mean": 0,
                        "count": 0,
                        "median": 0,
                    }
                return {
                    "max": round(mil_ds.describe[key]["max"], rounding),
                    "min": round(mil_ds.describe[key]["min"], rounding),
                    "std": round(mil_ds.describe[key]["std"], rounding),
                    "mean": round(mil_ds.describe[key]["mean"], rounding),
                    "count": round(mil_ds.describe[key]["count"], rounding),
                    # We don't have median in describe
                    "median": round(mil_ds.describe[key]["mean"], rounding),
                }
            step_count_walk = extract_result("stepCountWalk")["max"]
            step_count_run = extract_result("stepCountRun")["max"]
            step_count_ratio = round(
                step_count_walk / step_count_run, rounding
            ) if step_count_run > 0 else 0.0
            output_source = extract_result("outputSource")

            single_slot["result"] = {
                "speed": extract_result("gpsRecoverySpeed"),
                "distance": extract_result("gpsRecoveryDistance")["max"],
                "power": extract_result("gpsRecoveryPower"),
                "cadence": extract_result("cadence"),
                "undulation": extract_result("undulation"),
                "stiffness": extract_result("stiffness"),
                "stride_length": extract_result("strideLength"),
                "effective_flight_time":
                    extract_result("effectiveFlightTime"),
                "effective_contact_time":
                    extract_result("effectiveContactTime"),
                "gps_valid": bool(
                    (output_source is not None) and
                    (output_source["min"] == 4.0) and
                    (output_source["max"] == 4.0)),
                "regularity_90": 1,
                "step_count_ratio": step_count_ratio,
                "regularity_median": 1
            }
        return [single_slot]

    def post_resolve_slots(self, slots):
        for s in slots:
            s.workout = self

    def _get_slots(self, tags):
        return [x for x in self.slots if
                x.tags and set(tags).issubset(set(x.tags))]

    def get_slot_with_tags(self, tags):
        slots = self._get_slots(tags)
        if len(slots) > 0:
            return slots[0]
        else:
            return None

    def get_all_slots_with_tags(self, tags):
        return self._get_slots(tags)

    def dataset_names(self):
        return [x.name for x in self.datasets]

    def dataset_types(self):
        return set([x.type for x in self.datasets])


class Profile(WrapperObject):
    id = IDField()
    full_name = StringField()
    gender = StringField()
    weight = FloatField()
    height = FloatField()
    leg_length = FloatField()
    waist = FloatField()
    vma = FloatField()
    pma = FloatField()
    vo2_max = FloatField()
    birth_date = DateField()
    age = FloatField()
    running_style_score = FloatField()
    aerobic_capacity_score = FloatField()
    muscular_strength_score = FloatField()
    body_composition_score = FloatField()
    running_fitness_score = FloatField()

    def resolve_age(self, json, field):
        return relativedelta(date.today(), self.birth_date).years



import os

version_py = os.path.join(os.path.dirname(__file__), '../VERSION.txt')

try:
    with open(version_py, 'r') as fh:
        version_txt = open(
            version_py).read().strip().split('=')[-1].replace('"', '').strip()
except Exception:
    version_txt = "dev"
VERSION = version_txt



import abc
import requests
import fastavro as avro
import logging

from io import BytesIO

from myotest.version import VERSION
from myotest.exception import ClientError, ResourceNotFoundException
from myotest.models import Workout, Profile

logger = logging.getLogger("myotest")


USER_AGENT_VERISON = "python-revo-client/{}".format(VERSION)
PRODUCTION_URL = "https://api.myotest.cloud"
STAGING_URL = "https://api-staging.myotest.cloud"
DEV_URL = "http://localhost:8000"


class Client(object):

    def __init__(self):
        super().__init__()
        self.user_id = None

    @abc.abstractmethod
    def test(self):
        pass

    @abc.abstractmethod
    def fetch_resource(self, path, query):
        pass

    @abc.abstractmethod
    def fetch_avro(self, path):
        pass

    def get_resource(self, path, query, wrapper_class=None):
        return self._wrap_object(
            self.fetch_resource(path, query),
            wrapper_class=wrapper_class
        )

    def _wrap_object(self, json, wrapper_class=None):
        if isinstance(json, dict):
            return wrapper_class(client=self, json=json)
        elif isinstance(json, list):
            return [self._wrap_object(x, wrapper_class=wrapper_class)
                    for x in json]
        elif json is None:
            return None
        raise ValueError("Unknown type to wrap")

    def _workout_query(self, query={}):
        base = dict()
        if self.user_id:
            base["user"] = self.user_id
        else:
            base["user"] = self.get_profile().id
        return {**base, **query}

    def get_last_workout_with_tags(self, tags, at=None):
        result = self.fetch_resource(
            "/api/v1/workouts/", self._workout_query({"tag": tags, "at": at}))
        if result["count"] > 0:
            return self._wrap_object(result["results"][0], Workout)
        return None

    def get_workout_with_id(self, workout_id):
        return self.get_resource(
            "/api/v1/workouts/{}/".format(
                workout_id), self._workout_query(), Workout)

    def get_last_workouts(self, limit=20, at=None):
        result = self.fetch_resource(
            "/api/v1/workouts/",
            self._workout_query({"limit": limit, "at": at}))
        if result["count"] > 0:
            return self._wrap_object(result["results"], Workout)
        return []

    def get_profile(self, reload=False):
        if not hasattr(self, "_profile") or reload:
            if self.user_id:
                profile = self.get_resource(
                    "/api/profiles/{}/".format(self.user_id), {}, Profile)
            else:
                profile = self.get_resource(
                    "/api/profile/", {}, Profile)
            self._profile = profile
        return self._profile

    def get_slots(self, workout_id):
        slots_json = self.fetch_resource(
            "/api/workout-validation/{}/".format(workout_id), {})
        if slots_json:
            slot_info = {slot["id"]: slot for slot in slots_json["slots"]}
            return [
                {**slot_info[s["id"]], **s}
                for s in slots_json["validations"]["slots"]]
        else:
            return []


class RemoteClient(Client):
    def __init__(self, token, server=None, user_id=None):
        super().__init__()
        assert token is not None

        if server == "staging":
            self.url = STAGING_URL
        elif server == "production":
            self.url = PRODUCTION_URL
        elif server == "dev":
            self.url = DEV_URL
        else:
            self.url = server or PRODUCTION_URL

        self.authorization = "Token {}".format(token)
        self.user_id = user_id

    def test(self):
        result = self.fetch_resource("/api/monitor/")
        if isinstance(result, dict):
            return {
                "client-version": VERSION,
                "storage": result["storage"]
            }

    def get_headers(self):
        return {
            "user-agent": USER_AGENT_VERISON,
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": self.build_authorization()
        }

    def build_authorization(self):
        return self.authorization

    def fetch_resource(self, path, query={}):
        url = "{}{}".format(self.url, path)
        r = requests.get(url, headers=self.get_headers(), params=query)
        if r.status_code == 200:
            return r.json()
        elif r.status_code == 404:
            raise ResourceNotFoundException(path)
        else:
            raise ClientError(r.text, code=r.status_code)

    def fetch_avro(self, avro_url):
        logger.warning(
            "Downloading avro file, this should be avoided in production")
        response = requests.get(avro_url)
        if response.status_code == 200:
            return avro.reader(BytesIO(response.content))
        else:
            raise ClientError(response.text, code=response.status_code)



import unittest
import dateutil.parser
import mock
import os

from datetime import date, timedelta
from dateutil.relativedelta import relativedelta

from tests import load_json, load_content
from myotest import Client
from myotest.models import ResultData


class MockResponse(object):

    def __init__(self, content_path, code=200, url=None):
        self.status_code = code
        self.file_name = os.path.join(
            '_data', content_path)
        self.url = url

    def json(self):
        return load_json(self.file_name)

    @property
    def content(self):
        return load_content(self.file_name)

    @property
    def text(self):
        return load_content(self.file_name).decode("utf-8")


def requests_side_effects(path, headers={}, params={}):
    mapping = {
        "https://api.myotest.cloud/api/v1/workouts/":
            MockResponse("workout-test.json"),
        "https://test.com/data/001.avro":
            MockResponse("slot-1.avro"),
        "https://test.com/data/002.avro":
            MockResponse("mil-1.avro"),
        "https://api.myotest.cloud/api/workout-validation/"
        "e2fe9005-8d86-418e-976b-634ac374ff1e/":
            MockResponse("workout-validation-test.json"),
        "https://api.myotest.cloud/api/profiles/profile-999/":
            MockResponse("profile-999.json"),
        "https://api.myotest.cloud/api/profile/":
            MockResponse("profile-007.json"),
        "https://api.myotest.cloud/api/monitor/":
            MockResponse("monitor.json")
    }
    if path not in mapping:
        return MockResponse("404", code=404, url=path)
    return mapping[path]


def requests_side_effects_no_slots(path, headers={}, params={}):
    mapping = {
        "https://api.myotest.cloud/api/v1/workouts/":
            MockResponse("workout-test-no-slots.json"),
        "https://test.com/data/001.avro":
            MockResponse("slot-1.avro"),
        "https://test.com/data/002.avro":
            MockResponse("mil-1.avro"),
        "https://api.myotest.cloud/api/profile/":
            MockResponse("profile-007.json")
    }
    if path not in mapping:
        return MockResponse("404", code=404, url=path)
    return mapping[path]


def requests_side_effects_all_workouts(path, headers={}, params={}):
    mapping = {
        "https://api.myotest.cloud/api/v1/workouts/":
            MockResponse("all-workouts-test.json"),
        "https://api.myotest.cloud/api/profile/":
            MockResponse("profile-007.json")
    }
    if path not in mapping:
        return MockResponse("404", code=404, url=path)
    return mapping[path]


class ClientTest(unittest.TestCase):

    @mock.patch('requests.get', mock.Mock(side_effect=requests_side_effects))
    def test_base(self):
        client = Client("token-001")
        client.test()

    @mock.patch('requests.get', mock.Mock(side_effect=requests_side_effects))
    def test_get_workout(self):
        client = Client("token-001")
        workout = client.get_last_workout_with_tags(["test"])
        self.assertEqual(workout.title, "TEST workout 2017-10-06")
        self.assertEqual(workout.type, "data.open")
        self.assertEqual(
            workout.start,
            dateutil.parser.parse("2017-10-06T09:15:55.144470Z"))
        self.assertEqual(
            workout.end,
            dateutil.parser.parse("2017-10-06T09:30:55.144470Z"))
        self.assertEqual(
            workout.target_duration,
            timedelta(seconds=10)
        )
        self.assertEqual(
            workout.effective_duration,
            timedelta(seconds=900)
        )
        datasets = workout.datasets
        self.assertEqual(len(datasets), 3)

        names = set(map(lambda x: x.name, datasets))
        self.assertSetEqual(names, {"mil-1", "gps-1", "slot-1"})
        self.assertEqual(
            sorted(workout.dataset_names()), ["gps-1", "mil-1", "slot-1"])
        self.assertSetEqual(workout.dataset_types(), {"mil", "gps", "slot"})

    @mock.patch(
        'requests.get',
        mock.Mock(side_effect=requests_side_effects_all_workouts))
    def test_get_workouts(self):
        client = Client("token-001")
        workouts = client.get_last_workouts()
        self.assertEqual(len(workouts), 2)

    @mock.patch('requests.get', mock.Mock(side_effect=requests_side_effects))
    def test_get_workout_datasets(self):
        client = Client("token-001")
        workout = client.get_last_workout_with_tags(["test"])

        datasets = workout.get_datasets("mil")
        self.assertEqual(len(datasets), 1)
        self.assertEqual(datasets[0].name, "mil-1")

        datasets = workout.get_datasets("hot")
        self.assertEqual(len(datasets), 0)

        dataset = workout.get_dataset("mil")
        self.assertEqual(dataset.name, "mil-1")

        dataset = workout.get_dataset("mil-1")
        self.assertEqual(dataset.name, "mil-1")

        dataset = workout.get_dataset("slot")
        self.assertEqual(dataset.name, "slot-1")

        self.assertListEqual(
            list(dataset.dataframe.keys()),
            ['event', 'slot_id', 'time'])
        self.assertEqual(dataset.workout, workout)

        self.assertDictEqual(
            dataset.describe["time"],
            {
                '25%': 300.4151147156954,
                '50%': 450.3213799893856,
                '75%': 600.225209236145,
                'count': 6,
                'max': 900.4051049947739,
                'mean': 450.2991078197956,
                'min': 0.1072700023651123,
                'std': 314.6905086959979
            })
        self.assertListEqual(
            dataset.avro_schema.fields,
            [{'name': 'time', 'type': 'double'},
             {'name': 'event', 'type': {
                'name': 'trainingMakerType', 'type': 'enum',
                'symbols': ['SLOT_START', 'SLOT_END']}},
             {'name': 'slot_id', 'type': 'string'}])

    @mock.patch('requests.get', mock.Mock(side_effect=requests_side_effects))
    def test_get_workout_slots(self):
        client = Client("token-001")
        workout = client.get_last_workout_with_tags(["test", "other"])

        slots = workout.slots
        slot_by_id = {s.id: s for s in slots}

        slot = slot_by_id["c7dd04fd-94a0-9bdd-5d3f-a54f10b43021"]
        self.assertEqual(slot.tags, [
            "max_running_ranges",
            "vma_high"
        ])
        self.assertEqual(slot.text, "MAX 5 minutes")
        self.assertEqual(slot.type, "run")
        self.assertEqual(slot.start_time, timedelta(seconds=600.228))
        self.assertEqual(slot.end_time, timedelta(seconds=900.405))
        self.assertEqual(slot.value.type, "duration")
        self.assertEqual(slot.value.value, 300)

        self.assertEqual(slot.result.power.max, 1)
        self.assertEqual(slot.result.power.min, 0)
        self.assertEqual(slot.result.power.std, 0.5)
        self.assertEqual(slot.result.power.mean, 0.512)
        self.assertEqual(slot.result.power.median, 1.0)
        self.assertEqual(slot.result.power.count, 1000)

        self.assertIsInstance(slot.result.speed, ResultData)
        self.assertIsInstance(slot.result.cadence, ResultData)
        self.assertIsInstance(slot.result.undulation, ResultData)
        self.assertIsInstance(slot.result.stiffness, ResultData)
        self.assertIsInstance(slot.result.stride_length, ResultData)
        self.assertIsInstance(slot.result.effective_flight_time, ResultData)
        self.assertIsInstance(slot.result.effective_contact_time, ResultData)

        slot = workout.get_slot_with_tags(["vma_high"])
        self.assertEqual(slot.id, "c7dd04fd-94a0-9bdd-5d3f-a54f10b43021")
        mil_slot_df = slot.get_dataframe("mil")
        self.assertTrue(
            mil_slot_df.iloc[0]["time"] >= slot.start_time.total_seconds())
        self.assertTrue(
            mil_slot_df.iloc[0]["time"] <= slot.end_time.total_seconds())

    @mock.patch('requests.get', mock.Mock(side_effect=requests_side_effects))
    def test_get_profile(self):
        client = Client("token-001")
        profile = client.get_profile()
        self.assertEqual(profile.full_name, "James Bond")
        self.assertEqual(
            profile.birth_date, dateutil.parser.parse("1972-07-02").date())
        age = relativedelta(
            date.today(),
            dateutil.parser.parse("1972-07-02").date()).years
        self.assertEqual(
            profile.age, age)

    @mock.patch('requests.get', mock.Mock(side_effect=requests_side_effects))
    def test_get_other_profile(self):
        client = Client("token-001", user_id="profile-999")
        profile = client.get_profile()
        self.assertEqual(profile.full_name, "Bob Misterio")

    @mock.patch('requests.get', mock.Mock(
        side_effect=requests_side_effects_no_slots))
    def test_get_workout_no_slots(self):
        client = Client("token-001")
        workout = client.get_last_workout_with_tags(["test", "other"])

        slots = workout.slots
        self.assertEqual(len(slots), 1)
        slot = slots[0]
        self.assertEqual(slot.tags, [
            "max_running_ranges",
            "casual_running_ranges",
            "warm_up_running_ranges",
            "vma_high"
        ])
        self.assertEqual(slot.text, workout.title)
        self.assertEqual(slot.type, "unknown")
        self.assertEqual(slot.start_time, timedelta(seconds=0))
        self.assertEqual(slot.end_time, timedelta(seconds=900))
        self.assertEqual(slot.value.type, "duration")
        self.assertEqual(slot.value.value, 10.0)

        self.assertEqual(slot.result.speed.count, 3001)
        self.assertEqual(slot.result.speed.min, 2.798)
        self.assertEqual(slot.result.speed.max, 6.719)
        self.assertEqual(slot.result.speed.std, 1.516)
        self.assertEqual(slot.result.speed.mean, 5.068)
        self.assertEqual(slot.result.speed.median, 5.068)





class ClientError(Exception):
    def __init__(self, message, code=0, *args):
        super().__init__(message, *args)
        self.code = code


class ResourceNotFoundException(ClientError):
    def __init__(self, message, *args):
        super().__init__(message, code=404, *args)



# -*- coding: utf-8 -*-
import os
import subprocess

from distutils.core import setup
from setuptools import find_packages

with open('README.md') as readme:
    long_description = readme.read()

with open('requirements.txt') as reqs:
    install_requires = [
        line for line in reqs.read().split('\n') if (line and not
                                                     line.startswith('--'))
    ]

# Fetch version from git tags, and write to version.py.
# Also, when git is not available (PyPi package), use stored version.py.
version_py = os.path.join(os.path.dirname(__file__), 'VERSION.txt')

try:
    version_git = subprocess.check_output(
        ["git", "describe", "--tags"]).rstrip().decode('utf-8')
except Exception:
    with open(version_py, 'r') as fh:
        version_git = open(
            version_py).read().strip().split('=')[-1].replace('"', '').strip()

version_msg = "# Do not edit this file, " \
              "pipeline versioning is governed by git tags"
with open(version_py, 'w') as fh:
    fh.write(version_msg + os.linesep + 'VERSION=' + version_git)

setup(
    name='revo-client',
    version=version_git,
    author='myotest SA',
    author_email='dev@myotest.com',
    packages=find_packages(),
    include_package_data=True,
    url='https://github.com/Myotest/python-revo-client',
    license='MIT',
    description='Python client for Myotest smart coaching API',
    long_description=long_description,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Pike Modules',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    zip_safe=False,
    install_requires=install_requires,
    test_suite='tests',
    python_requires='>=3',
)



import dateutil.parser
import json
from datetime import timedelta


class Field(object):

    def __init__(self, source=None):
        self.source = source
        self.client = None

    def convert(self, value, client):
        return value


class StringField(Field):

    def convert(self, value, client):
        return str(value) if value else None


class BooleanField(Field):

    def convert(self, value, client):
        return True if value else False


class IDField(StringField):
    pass


class FloatField(Field):

    def convert(self, value, client):
        return float(value) if value is not None else None


class IntegerField(Field):

    def convert(self, value, client):
        return int(value) if value is not None else None


class DateTimeField(Field):

    def convert(self, value, client):
        return dateutil.parser.parse(value) if value else None


class DateField(Field):

    def convert(self, value, client):
        return dateutil.parser.parse(value).date() if value else None


class RawField(Field):

    def convert(self, value, client):
        return value


class DictField(Field):

    def convert(self, value, client):
        if value is None:
            return None
        assert isinstance(value, dict)
        return value


class DurationField(Field):

    def convert(self, value, client):
        return timedelta(seconds=float(value))


class DataframeField(Field):

    def convert(self, value, client):
        return value


class LinkField(Field):

    def convert(self, value, client):
        return value


class ListField(Field):

    def __init__(self, subfield: Field, source=None):
        super().__init__(source=source)
        assert subfield is not None
        self.subfield = subfield

    def convert(self, value, client):
        if not value:
            return None
        elif isinstance(value, list):
            return list(map(
                lambda x: self.subfield.convert(x, client),
                value
            ))
        else:
            raise ValueError("Value is not a list")


class WrapperObject(Field):

    def __init__(self, client=None, json=None):
        super().__init__()
        if json and not isinstance(json, dict) and not isinstance(json, list):
            raise ValueError("json should nbe a dict or list, got {}".format(
                json
            ))
        self.client = client
        self._json = json
        self._cache = dict()

    def __getattribute__(self, name):
        json = super().__getattribute__("_json")
        cls = super().__getattribute__("__class__")
        cache = super().__getattribute__("_cache")
        client = super().__getattribute__("client")

        if not client:
            return super().__getattribute__(name)

        if name in cache:
            # Cache already contains value
            return cache[name]

        field = getattr(cls, name, None)
        if isinstance(field, Field):
            if name in cache:
                return cache[name]
            json_name = field.source or name
            resolver_fun = getattr(self, "resolve_{}".format(name), None)
            if resolver_fun:
                # we have a resolve function
                orig = resolver_fun(json, field)
            elif isinstance(field, LinkField):
                # Special case to create cyclic references
                return super().__getattribute__(name)
            else:
                orig = json.get(json_name, None)
            if isinstance(orig, WrapperObject):
                # already wrapped
                target = orig
            else:
                target = field.convert(orig, self.client)

            post_resolver_fun = getattr(
                self, "post_resolve_{}".format(name), None)
            if post_resolver_fun:
                new_target = post_resolver_fun(target)
                if new_target is not None:
                    target = new_target

            cache[name] = target
            return target
        else:
            return super().__getattribute__(name)

    def json(self):
        return self._json

    def convert(self, value, client):
        # Duplicate object with filled client and value
        return self.__class__(client, value)

    def __repr__(self):
        if not self.client:
            return "<Prototype: {}>".format(self.__class__)
        else:
            return "<Wrapper [{}]: {}>".format(
                self.__class__,
                json.dumps(self.json(), indent=4))
